# include <stdio.h>
# include "parm.h"
# include "stree/ststructs.mh"
# include "datatypes/stack.h"
# include "stree/is_ptr.h"

#/*
# *  Insert the links comprising the symbol table into the syntax tree.
# */

def build_symtab(x)
  $stderr.puts "build_symtab x = " + x.to_s + "\n" 
  if (x == RNIL) then
    # /* Ignore missing clauses.*/
    return
  end
  case ( x.kind ) 
  when LISTHEADER
    # /* Process every element on the list.*/ frag000005
    #    maplist(v, x, build_symtab(v))
    # 1 "in/frag000005.rb.c"
    mL__0O = ((x).lh_first)
    while(mL__0O != RNIL) do
      v = cn_head(mL__0O)
      build_symtab(v) 
      mL__0O = cn_tail(mL__0O) 
    end
    #    break
  when DECLARATION  # /* Only for program body declaration.*/
    # /* Add id to symtab.*/
    add_id(x, x)
    # /* Process program body.Leave decl_block pointing to RNIL.*/
    build_symtab(x.decl_denotation)
    #            break
  when PARAMETER
    # /* Should never get here.*/
    bs_error(x)
  when RECORDELEMENT
    # /* Should never get here.*/
    bs_error(x)
  when VARSIGNATURE
    # /* Process denotation.*/
    build_symtab(x.var_denotation)
    # break
  when VALSIGNATURE
    # /* Process denotation.*/
    build_symtab(x.val_denotation)
    #            break
  when FUNCSIGNATURE
    # /* Add the parameter identifiers to symtab.*/ frag000006
    #    maplist(v, x.fsig_param_list, add_id(v, x))
    # 1 "in/frag000006.rb.c"
    mL__0O = ((x.fsig_param_list).lh_first) 
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O)
      add_id(v, x)
      mL__0O = cn_tail(mL__0O) 
    end
    # /* Process the parameter signatures and the function */
    # /* result signature.                                */ frag000007
    #    maplist(v, x.fsig_param_list, build_symtab(v.par_signature))
    # 1 "in/frag000007.rb.c"
    mL__0O = ((x.fsig_param_list).lh_first) 
    while(mL__0O != RNIL) do
      v = cn_head(mL__0O) 
      build_symtab(v.par_signature)
      mL__0O = cn_tail(mL__0O) 
    end
    build_symtab(x.fsig_result_sig)
    # /* Restore surrounding scope.*/ frag000008
    #    maplist(v, x.fsig_param_list, restore(v))
    # 1 "in/frag000008.rb.c"
    mL__0O = ((x.fsig_param_list).lh_first)
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O) 
      restore(v)
      mL__0O = cn_tail(mL__0O) 
    end
    #        break
  when TYPESIGNATURE
    # /* Add local id to symtab.*/
    add_id(x, x)
    # /* push(x,lti_stack) */
    # /* Process each component signature.*/
    # frag000004
    # 1 "in/frag000004.rb.c"
    mL__0O = x.ts_clist.lh_first 
    while (mL__0O != RNIL) do  
      v = cn_head(mL__0O) 
      case (v.kind) 
      when TSCOMPONENT 
        build_symtab(v.tsc_signature) 
        #  break 
      when DEFCHARSIGS 
        #break 
      else 
        bs_error(v)
      end 
      mL__0O = cn_tail(mL__0O) 
    end
    # 13 "in/frag000004.rb.c"
    
    # /* Restore surrounding scope.*/
    restore(x)
    # /* pop(lti_stack) */
    # break
  when TSCOMPONENT
    # /* Should never get here.*/
    bs_error(x)
  when BLOCKDENOTATION
    # /* Add declaration identifiers to symtab.*/ frag000009
    #      maplist(v, x.bld_declaration_list, add_id(v, x))
    # 1 "in/frag000009.rb.c"
    
    mL__0O = ((x.bld_declaration_list).lh_first)
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O) 
      add_id(v, x)
      mL__0O = cn_tail(mL__0O) 
    end 
    # /* Process declaration signatures, declaration denotations, */
    # /* and the block denotation.        frag000010                       */
    #                maplist(v, x.bld_declaration_list, 
    #                        { build_symtab(v.decl_signature)
    #                          build_symtab(v.decl_denotation) })
    # 1 "in/frag000010.rb.c"
    mL__0O = ((x.bld_declaration_list).lh_first) 
    while (mL__0O != RNIL) do
      v =  cn_head(mL__0O)
      build_symtab(v.decl_signature) 
      build_symtab(v.decl_denotation)
      mL__0O = cn_tail(mL__0O) 
    end
    build_symtab(x.bld_den_seq)
    
    # /* Restore surrounding scope.*/ frag000011
    #                maplist(v, x.bld_declaration_list, restore(v))
    # 1 "in/frag000011.rb.c"
    mL__0O = ((x.bld_declaration_list).lh_first)
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O)
      restore(v)
      mL__0O = cn_tail(mL__0O) 
    end
    #            break
  when USELIST
    x.usl_previous_list = usl_ptr
    # /* Process list of types with old use list.*/
    build_symtab(x.usl_type_list)
    # /* Process body with new use list */
    usl_ptr = x
    build_symtab(x.usl_den_seq)
    # /* restore old usl_ptr value */
    usl_ptr = x.usl_previous_list
    #   break
  when APPLICATION
    # /* Process operator and args.*/
    build_symtab(x.ap_operator)
    build_symtab(x.ap_args)
    #  break
  when ENUMERATION   # frag000012
    #	    maplist(v, x.enum_id_list, {
    #		x.id_last_definition = x
    #	    })
    # 1 "in/frag000012.rb.c"
    mL__0O = ((x.enum_id_list).lh_first) 
    while(mL__0O != RNIL) do
      v = cn_head(mL__0O)
      x.id_last_definition = x
      mL__0O = cn_tail(mL__0O) 
    end
    #   break
  when EXTENSION
    # /* Process denotation.*/
    build_symtab(x.ext_denotation)
    # break
  when UNIONCONSTRUCTION
    # /* Add local type name to symtab.*/
    add_id(x, x)
    if (x.kind == PRODCONSTRUCTION) then
      # /* Add field names to symtab.*/ frag000013
      #		maplist(v, x.prod_components,
      #			add_id(v,x))
      # 1 "in/frag000013.rb.c"
      mL__0O = ((x.prod_components).lh_first) 
      while (mL__0O != RNIL) do
        v = cn_head(mL__0O)
        add_id(v,x)
        mL__0O = cn_tail(mL__0O) 
      end
    end      
    # /* Process the component signatures */ frag000014
    #                maplist(v, x.prod_components,
    #                        build_symtab(v.par_signature))
    # 1 "in/frag000014.rb.c"
    mL__0O = (x.prod_components).lh_first
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O)
      build_symtab(v.par_signature)
      mL__0O = cn_tail(mL__0O) 
    end
    # /* Restore surrounding scope.*/
    restore(x)
    if (x.kind == PRODCONSTRUCTION) then  # frag000015
      #		  maplist(v, x.prod_components,
      #			  restore(v))
      # 1 "in/frag000015.rb.c"
      mL__0O = ((x.prod_components).lh_first) 
      while (mL__0O != RNIL) do 
        v =  cn_head(mL__0O) 
        restore(v)
        mL__0O = cn_tail(mL__0O) 
      end
    end
  when PRODCONSTRUCTION
    # /* Add local type name to symtab.*/
    add_id(x, x)
    if (x.kind == PRODCONSTRUCTION) then
      # /* Add field names to symtab.*/ frag000013
      #		maplist(v, x.prod_components,
      #			add_id(v,x))
      # 1 "in/frag000013.rb.c"
      mL__0O = ((x.prod_components).lh_first) 
      while (mL__0O != RNIL) do
        v = cn_head(mL__0O)
        add_id(v,x)
        mL__0O = cn_tail(mL__0O) 
      end
    end      
    # /* Process the component signatures */ frag000014
    #                maplist(v, x.prod_components,
    #                        build_symtab(v.par_signature))
    # 1 "in/frag000014.rb.c"
    mL__0O = (x.prod_components).lh_first
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O)
      build_symtab(v.par_signature)
      mL__0O = cn_tail(mL__0O) 
    end
    # /* Restore surrounding scope.*/
    restore(x)
    if (x.kind == PRODCONSTRUCTION) then  # frag000015
      #		  maplist(v, x.prod_components,
      #			  restore(v))
      # 1 "in/frag000015.rb.c"
      mL__0O = ((x.prod_components).lh_first) 
      while (mL__0O != RNIL) do 
        v =  cn_head(mL__0O) 
        restore(v)
        mL__0O = cn_tail(mL__0O) 
      end
    end
    #            break
  when RECORDCONSTRUCTION
    # /* Process record component denotations.*/ frag000016
    #                maplist(v, x.rec_component_list, {
    #                    v.re_id.id_last_definition = v.re_id
    #		    build_symtab(v.re_denotation)
    #		})
    # 1 "in/frag000016.rb.c"
    mL__0O = ((x.rec_component_list).lh_first) 
    while (mL__0O != RNIL) do
      v =  cn_head(mL__0O) 
      v.re_id.id_last_definition = v.re_id 
      build_symtab(v.re_denotation)
      mL__0O = cn_tail(mL__0O) 
    end
    #      break
  when WITHLIST
    # /* Process declaration denotations.*/ frag000017
    #                maplist(v, x.wl_component_list,
    #			{ 
    #                          v.decl_id.id_last_definition = RNIL
    #                          build_symtab(v.decl_signature)
    #                          build_symtab(v.decl_denotation)
    #                       } )
    # 1 "in/frag000017.rb.c"
    mL__0O = ((x.wl_component_list).lh_first) 
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O) 
      v.decl_id.id_last_definition = RNIL
      build_symtab(v.decl_signature) 
      build_symtab(v.decl_denotation)
      mL__0O = cn_tail(mL__0O) 
    end
    #            break
  when MODPRIMARY
    # /* Process primary and modifier.*/
    build_symtab(x.mp_primary)
    # /* add local id to symtab */
    add_id(x,x)
    build_symtab(x.mp_type_modifier)
    # /* restore surrounding scope */
    restore(x)
    #        break
  when EXPORTLIST
    build_symtab(x.el_export_element_list)
  when HIDELIST
    # /* Process the export element list.*/
    build_symtab(x.el_export_element_list)
    #    break
  when EXPORTELEMENT
    x.ee_id.id_last_definition = x.ee_id
    build_symtab(x.ee_signature)
    build_symtab(x.ee_export_list)
    # break
  when ALLCONSTANTS
    # /* Nothing to do.*/
    #      break
  when WORDELSE
    # /* Nothing to do.*/
    # break
  when WORDCAND
    # /* Nothing to do.*/
    # break
  when WORDCOR
    # /* Nothing to do.*/
    # break
  when GUARDEDLIST
    # /* Process guarded list.*/
    build_symtab(x.gl_list)
    # break
  when LOOPDENOTATION
    # /* Process guarded list.*/
    build_symtab(x.gl_list)
    # break
  when GUARDEDELEMENT
    # /* Process guard and element.*/
    build_symtab(x.ge_guard)
    build_symtab(x.ge_element)
    #        break
  when OPRID  # /* Just a use, not a definition.*/
    # /* Process selection type and signature */
    build_symtab(x.sel_type)
    build_symtab(x.signature)
    # /* Fill in last definition field.*/
    process_id(x)
  when LETTERID
    # /* Process selection type and signature */
    build_symtab(x.sel_type)
    build_symtab(x.signature)
    # /* Fill in last definition field.*/
    process_id(x)
    # break
  when QSTR
    # /* process selection type */
    build_symtab(x.sel_type)
    
    # /* fill in use list field */
    x.str_use_list = usl_ptr
  when UQSTR
    # /* process selection type */
    build_symtab(x.sel_type)
    
    # /* fill in use list field */
    x.str_use_list = usl_ptr
    # break
  when FUNCCONSTR
    # /* Add the parameter identifiers to symtab.*/
    #               ifdef DEBUG
    if(!is_ptr(x.signature) || x.signature == RNIL) then
      dbgmsg("build_symtab bad fn signature\n")
      abort()
    end
    #               endif  frag000018
    #		maplist(v, x.signature.fsig_param_list, add_id(v, x))
    # 1 "in/frag000018.rb.c"
    mL__0O = ((x.signature.fsig_param_list).lh_first) 
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O) 
      add_id(v, x)
      mL__0O = cn_tail(mL__0O)  
    end        
    # /* Process the parameter signatures, the function result */
    # /* signature, and the function body.frag000019                     */
    #		maplist(v, x.signature.fsig_param_list,
    #			build_symtab(v.par_signature))
    # 1 "in/frag000019.rb.c"
    mL__0O = ((x.signature.fsig_param_list).lh_first) 
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O)
      build_symtab(v.par_signature)
      mL__0O = cn_tail(mL__0O)  
    end          
    build_symtab(x.signature.fsig_result_sig)
    build_symtab(x.fc_body)
    # /* Restore surrounding scope.*/  frag000020
    #                maplist(v, x.signature.fsig_param_list, restore(v))
    # 1 "in/frag000020.rb.c"
    mL__0O = ((x.signature.fsig_param_list).lh_first) 
    while (mL__0O != RNIL) do
      v = cn_head(mL__0O) 
      restore(v)
      mL__0O = cn_tail(mL__0O) 
    end
    #	    break
  when REXTERNDEF
  when EXTERNDEF
    #	    break
  end
end
#DJDSTOP
extern add_id(), restore();

/*   Stack of enclosing type signature nodes.
 * Used in determining the defining nodes
 * for identifiers with a -1 string table index.
 * (See streedefs.h.) 
 * NOT NEEDED - see process_id.c
   STACK lti_stack = emptystack();
 */

/*  Pointer to next surrounding use list. Used  */
/* to set fields in identifier, string and       */
/* use list nodes.                              */
NODE * usl_ptr = RNIL;


/*
 *  Insert the links comprising the symbol table into the syntax tree.
 */

build_symtab(x)
NODE * x;
{

    if (x == NIL)
        /* Ignore missing clauses. */
            return;

    switch ( x -> kind ) {

        case LISTHEADER:
            /* Process every element on the list. */ 
                maplist(v, x, build_symtab(v));
            break;


        case DECLARATION: /* Only for program body declaration. */
            /* Add id to symtab. */
                add_id(x, x);

            /* Process program body. Leave decl_block pointing to NIL. */
                build_symtab(x->decl_denotation);
            break;


        case PARAMETER:
            /* Should never get here. */
                bs_error(x);


        case RECORDELEMENT:
            /* Should never get here. */
                bs_error(x);


        case VARSIGNATURE:
            /* Process denotation. */
                build_symtab(x->var_denotation);
            break;


        case VALSIGNATURE:
            /* Process denotation. */
                build_symtab(x->val_denotation);
            break;


        case FUNCSIGNATURE:
            /* Add the parameter identifiers to symtab. */
                maplist(v, x->fsig_param_list, add_id(v, x));

            /* Process the parameter signatures and the function */
            /* result signature.                                 */
                maplist(v, x->fsig_param_list,
                        build_symtab(v->par_signature));
                build_symtab(x->fsig_result_sig);

            /* Restore surrounding scope. */
                maplist(v, x->fsig_param_list, restore(v));
            break;
             

        case TYPESIGNATURE:
            /* Add local id to symtab. */
                add_id(x, x);
                /* push(x,lti_stack); */
            /* Process each component signature. */
                maplist(v, x->ts_clist,
                            switch (v->kind) {
				case TSCOMPONENT:
                                    /* v -> tsc_id -> id_last_definition = NIL; */
                                    build_symtab(v->tsc_signature);
                                    break;
                                case DEFCHARSIGS:
                                    /* Ignore. */
                                    break;
                                default:
                                    bs_error(v);
                            } );

            /* Restore surrounding scope. */
                restore(x);
                /* pop(lti_stack); */
            break;


        case TSCOMPONENT:
            /* Should never get here. */
                bs_error(x);


        case BLOCKDENOTATION:
            /* Add declaration identifiers to symtab. */
                maplist(v, x->bld_declaration_list, add_id(v, x));

            /* Process declaration signatures, declaration denotations, */
            /* and the block denotation.                                */
                maplist(v, x->bld_declaration_list,
                        { build_symtab(v->decl_signature);
                          build_symtab(v->decl_denotation); });
                build_symtab(x->bld_den_seq);

            /* Restore surrounding scope. */
                maplist(v, x->bld_declaration_list, restore(v));
            break;


        case USELIST:
            x->usl_previous_list = usl_ptr;
            /* Process list of types with old use list. */
                build_symtab(x->usl_type_list);
            /* Process body with new use list */
                usl_ptr = x;
                build_symtab(x->usl_den_seq);
            /* restore old usl_ptr value */
                usl_ptr = x->usl_previous_list;
            break;


        case APPLICATION:
            /* Process operator and args. */
                build_symtab(x->ap_operator);
                build_symtab(x->ap_args);
            break;


        case ENUMERATION:
	    maplist(v, x -> enum_id_list, {
		x -> id_last_definition = x;
	    });
            break;


        case EXTENSION:
            /* Process denotation. */
                build_symtab(x->ext_denotation);
            break;


        case UNIONCONSTRUCTION:
	case PRODCONSTRUCTION:
            /* Add local type name to symtab. */
		add_id(x, x);
	    if (x -> kind == PRODCONSTRUCTION) {
	      /* Add field names to symtab. */
		maplist(v, x -> prod_components,
			add_id(v,x));
	    }

            /* Process the component signatures */
                maplist(v, x->prod_components,
                        build_symtab(v->par_signature));

            /* Restore surrounding scope. */
                restore(x);
		if (x -> kind == PRODCONSTRUCTION) {
		  maplist(v, x -> prod_components,
			  restore(v));
		}

            break;
             

        case RECORDCONSTRUCTION:
            /* Process record component denotations. */
                maplist(v, x->rec_component_list, {
                    v -> re_id -> id_last_definition = v -> re_id;
		    build_symtab(v->re_denotation);
		});
            break;


        case WITHLIST:
            /* Process declaration denotations. */
                maplist(v, x->wl_component_list,
			{ 
                          v -> decl_id -> id_last_definition = NIL;
                          build_symtab(v->decl_signature);
                          build_symtab(v->decl_denotation);
                        } );

            break;


        case MODPRIMARY:
            /* Process primary and modifier. */
                build_symtab(x->mp_primary);

                /* add local id to symtab */
                    add_id(x,x);

                build_symtab(x->mp_type_modifier);

                /* restore surrounding scope */
                    restore(x);
            break;


        case EXPORTLIST:
        case HIDELIST:
            /* Process the export element list. */
                build_symtab(x->el_export_element_list);
            break;

        case EXPORTELEMENT:
	    x -> ee_id -> id_last_definition = x -> ee_id;
	    build_symtab(x -> ee_signature);
	    build_symtab(x -> ee_export_list);
            break;

        case ALLCONSTANTS:
            /* Nothing to do. */
            break;


        case WORDELSE:
            /* Nothing to do. */
            break;


        case WORDCAND:
            /* Nothing to do. */
            break;


        case WORDCOR:
            /* Nothing to do. */
            break;


        case GUARDEDLIST:
            /* Process guarded list. */
                build_symtab(x->gl_list);
            break;


        case LOOPDENOTATION:
            /* Process guarded list. */
                build_symtab(x->gl_list);
            break;


        case GUARDEDELEMENT:
            /* Process guard and element. */
                build_symtab(x->ge_guard);
                build_symtab(x->ge_element);
            break;


	case OPRID: /* Just a use, not a definition. */
	case LETTERID:
            /* Process selection type and signature */
                build_symtab(x->sel_type);
                build_symtab(x->signature);
            /* Fill in last definition field. */
                process_id(x);
            break;


	case QSTR:
	case UQSTR:
	    /* process selection type */
		build_symtab(x -> sel_type);

            /* fill in use list field */
                x->str_use_list = usl_ptr;
            break;


        case FUNCCONSTR:
	    /* Add the parameter identifiers to symtab. */
#               ifdef DEBUG
		  if(!is_ptr(x -> signature) || x -> signature == NIL) {
		    dbgmsg("build_symtab: bad fn signature\n");
		    abort();
		  }
#               endif
		maplist(v, x->signature->fsig_param_list, add_id(v, x));

            /* Process the parameter signatures, the function result */
	    /* signature, and the function body.                     */
		maplist(v, x->signature->fsig_param_list,
			build_symtab(v->par_signature));
		build_symtab(x->signature->fsig_result_sig);
		build_symtab(x->fc_body);

            /* Restore surrounding scope. */
                maplist(v, x->signature->fsig_param_list, restore(v));
	    break;

        case REXTERNDEF:
	case EXTERNDEF:
	    break;

    }
}


/*
 *  bs_error(x)
 *
 *  Print an error message and abort.
 */
bs_error(x)
NODE * x;
{
    dbgmsg("build_symtab: bad node: node addr = %o, kind = %s",
            x, kindname(x->kind));
    abort();
}


#DJDSTART
